/* 
 * Copyright (c) 1989 NeXT, Inc.
 *
 * HISTORY
 * 8-Jun-92  Matthew Self (mself) at NeXT
 *	Added mach_header field to kern_server_var_t for Objective-C.
 *
 * 12-Jul-90  Gregg Kellogg (gk) at NeXT
 *	Changed msgs to msg, as only a single message is retained.
 *	Added KS_VERSION(=2), which is updated whenever a protocol change
 *	is made.
 *
 * 51-May-90  Gregg Kellogg (gk) at NeXT
 *	Added pad filed at end of kern_server struct so that already compiled
 *	servers won't be impacted when the kern_server grows.
 *
 * 04-Apr-89  Gregg Kellogg (gk) at NeXT
 *	Created.
 *
 */

#ifndef	_KERN_SERVER_
#define	_KERN_SERVER_
#if	KERNEL
#import <mach/mach_user_internal.h>
#endif
#import <mach/mach_interface.h>
#import <mach/message.h>
#import <sys/types.h>
#import <kernserv/queue.h>
#import <kernserv/kern_server_reply_types.h>

/*
 * Messages to be sent to the Mach bootstrap port.  The bootstrap port
 * is inherited from the parent task and is used for finding information
 * about the external and internal environment.
 *
 * Generic things that need to be done to get a kernel server started:
 *	Wire down memory regions
 *	Associate ports with procedures (to handle messages)
 */

/*
 * This must be small enough to fit (along with a message header) in 8K.
 */
#define	KERN_SERVER_LOG_SIZE	500

#define KERN_SERVER_ERROR	100	
#define KERN_SERVER_NOTLOGGING	101	
#define KERN_SERVER_UNCREC_PORT	102	
#define KERN_SERVER_BAD_VERSION	103	

#define KERN_SERVER_NPORTPROC	50

typedef boolean_t (*port_map_proc_t) (
	msg_header_t *in_msg,
	void *arg);

typedef boolean_t (*port_map_serv_t) (
	msg_header_t *in_msg,
	msg_header_t *out_msg);

typedef boolean_t (*port_death_proc_t)(port_name_t port);
typedef boolean_t (*port_notify_proc_t)(port_name_t port, u_int notify_type);

typedef int (*call_proc_t)(int arg);

typedef struct port_proc_map {
	port_name_t	port;	
	port_map_proc_t	proc;	
	void *		uarg;
	enum { PP_handler, PP_server }
			type;	
} port_proc_map_t;

typedef struct {
	port_t		reply_port;
	port_name_t	req_port;
	queue_chain_t	link;
} ks_notify_t;

#if	KERNEL
/*
 * Supported compatibility level.
 */
#define	KS_COMPAT	2
#define	KS_VERSION	2

/*
 * Server instance structure
 */
typedef struct kern_server_var {
	simple_lock_data_t	slock;	
	port_name_t		local_port;
	port_name_t		task_port;
	thread_t		server_thread;
	port_name_t		bootstrap_port;
	port_name_t		boot_listener_port;
	port_name_t		log_port;
	port_name_t		notify_port;
	port_set_name_t		port_set;
	log_t			log;	
	queue_head_t		msg_callout_q;
	queue_head_t		msg_callout_fq;
	msg_header_t		*msg;	
	int			msg_size;
	struct msg_send_entry {
		void		(*func)();
		void *		arg;
		queue_chain_t	link;
	} msg_send_array[20];
	port_proc_map_t		port_proc[KERN_SERVER_NPORTPROC];
	port_name_t		last_unrec_port
	port_name_t		last_rec_port;
	int			last_rec_index;
	port_death_proc_t	pd_proc;
	port_notify_proc_t	pn_proc;
	queue_head_t		notify_q;
	int			version;
	port_t			kernel_port;
	struct mach_header	*mach_header;
} kern_server_var_t;

typedef kern_server_var_t *kern_server_t;

/*
 * Structure access functions.
 */
port_t kern_serv_local_port(kern_server_t *ksp);
port_t kern_serv_bootstrap_port(kern_server_t *ksp);
port_t kern_serv_notify_port(kern_server_t *ksp);
port_set_name_t kern_serv_port_set(kern_server_t *ksp);

/*
 * Function prototypes.
 */
void kern_server_main(void);
void kern_serv_port_gone(kern_server_t *ksp, port_name_t port);

/*
 * Log functions.
 */
void kern_serv_log (		
	kern_server_t	*ksp,	
	int		log_level,
	char		*msg,	
	int		arg1,
	int		arg2,
	int		arg3,
	int		arg4,
	int		arg5);

/*
 * Alias for msg_send to ensure that messages are sent at ipl0 from the proper
 * task.
 */
kern_return_t kern_serv_callout (
	kern_server_t	*ksp,
	void		(*proc)(void *arg),
	void		*arg);

/*
 * Kernel version of kern_serv_notify, doesn't contact kern_loader, uses
 * internal port_request_notification facility.
 */
kern_return_t kern_serv_notify (
	kern_server_t	*ksp,
	port_t		reply_port,
	port_t		req_port);

/*
 * Kernel reference for kern_serv interface routines.
 */
kern_return_t kern_serv_instance_loc (
	void		*arg,	
	vm_address_t	instance_loc);
kern_return_t kern_serv_boot_port (
	void		*arg,	
	port_t		boot_port);
kern_return_t kern_serv_version (
	void		*arg,	
	int		version);
kern_return_t kern_serv_load_objc (
	void		*arg,	
	vm_address_t	header);
kern_return_t kern_serv_wire_range (
	void		*arg,	
	vm_address_t	addr,
	vm_size_t	size);
kern_return_t kern_serv_unwire_range (
	void		*arg,	
	vm_address_t	addr,
	vm_size_t	size);
kern_return_t kern_serv_port_proc (
	void		*arg,	
	port_all_t	port,	
	port_map_proc_t	proc,	
	int		uarg);	
kern_return_t kern_serv_port_death_proc (
	void			*arg,
	port_death_proc_t	proc);
kern_return_t kern_serv_call_proc (
	void		*arg,	
	call_proc_t	proc,	
	int		uarg);	
kern_return_t kern_serv_shutdown (
	void		*arg);	
kern_return_t kern_serv_log_level (
	void		*arg,	
	int		log_level);
kern_return_t kern_serv_get_log (
	void		*arg,	
	port_t		reply_port);
kern_return_t kern_serv_port_serv (
	void		*arg,	
	port_all_t	port,	
	port_map_proc_t	proc,	
	int		uarg);	
port_t kern_serv_kernel_task_port(void);

#endif
#endif
